"""
TFIDF module
"""

import math
import os

from collections import Counter
from multiprocessing.pool import ThreadPool

import numpy as np

from ..pipeline import Tokenizer
from ..serialize import Serializer

from .base import Scoring
from .terms import Terms


class TFIDF(Scoring):
    """
    Term frequency-inverse document frequency (TF-IDF) scoring.
    """

    def __init__(self, config=None):
        super().__init__(config)

        # Document stats
        self.total = 0
        self.tokens = 0
        self.avgdl = 0

        # Word frequency
        self.docfreq = Counter()
        self.wordfreq = Counter()
        self.avgfreq = 0

        # IDF index
        self.idf = {}
        self.avgidf = 0

        # Tag boosting
        self.tags = Counter()

        # Tokenizer, lazily loaded as needed
        self.tokenizer = None

        # Term index
        self.terms = Terms(self.config["terms"], self.score, self.idf) if self.config.get("terms") else None

        # Document data
        self.documents = {} if self.config.get("content") else None

        # Normalize scores
        self.normalize = self.config.get("normalize")
        self.avgscore = None

    def insert(self, documents, index=None, checkpoint=None):
        # Insert documents, calculate word frequency, total tokens and total documents
        for uid, document, tags in documents:
            # Extract text, if necessary
            if isinstance(document, dict):
                document = document.get(self.text, document.get(self.object))

            if document is not None:
                # If index is passed, use indexid, otherwise use id
                uid = index if index is not None else uid

                # Add entry to index if the data type is accepted
                if isinstance(document, (str, list)):
                    # Store content
                    if self.documents is not None:
                        self.documents[uid] = document

                    # Convert to tokens, if necessary
                    tokens = self.tokenize(document) if isinstance(document, str) else document

                    # Add tokens for id to term index
                    if self.terms is not None:
                        self.terms.insert(uid, tokens)

                    # Add tokens and tags to stats
                    self.addstats(tokens, tags)

                # Increment index
                index = index + 1 if index is not None else None

    def delete(self, ids):
        # Delete from terms index
        if self.terms:
            self.terms.delete(ids)

        # Delete content
        if self.documents:
            for uid in ids:
                self.documents.pop(uid)

    def index(self, documents=None):
        # Call base method
        super().index(documents)

        # Build index if tokens parsed
        if self.wordfreq:
            # Calculate total token frequency
            self.tokens = sum(self.wordfreq.values())

            # Calculate average frequency per token
            self.avgfreq = self.tokens / len(self.wordfreq.values())

            # Calculate average document length in tokens
            self.avgdl = self.tokens / self.total

            # Compute IDF scores
            idfs = self.computeidf(np.array(list(self.docfreq.values())))
            for x, word in enumerate(self.docfreq):
                self.idf[word] = float(idfs[x])

            # Average IDF score per token
            self.avgidf = float(np.mean(idfs))

            # Calculate average score across index
            self.avgscore = self.score(self.avgfreq, self.avgidf, self.avgdl)

            # Filter for tags that appear in at least 1% of the documents
            self.tags = Counter({tag: number for tag, number in self.tags.items() if number >= self.total * 0.005})

        # Index terms, if available
        if self.terms:
            self.terms.index()

    def weights(self, tokens):
        # Document length
        length = len(tokens)

        # Calculate token counts
        freq = self.computefreq(tokens)
        freq = np.array([freq[token] for token in tokens])

        # Get idf scores
        idf = np.array([self.idf[token] if token in self.idf else self.avgidf for token in tokens])

        # Calculate score for each token, use as weight
        weights = self.score(freq, idf, length).tolist()

        # Boost weights of tag tokens to match the largest weight in the list
        if self.tags:
            tags = {token: self.tags[token] for token in tokens if token in self.tags}
            if tags:
                maxWeight = max(weights)
                maxTag = max(tags.values())

                weights = [max(maxWeight * (tags[tokens[x]] / maxTag), weight) if tokens[x] in tags else weight for x, weight in enumerate(weights)]

        return weights

    def search(self, query, limit=3):
        # Check if term index available
        if self.terms:
            # Parse query into terms
            query = self.tokenize(query) if isinstance(query, str) else query

            # Get topn term query matches
            scores = self.terms.search(query, limit)

            # Normalize scores, if enabled
            if self.normalize and scores:
                # Calculate max score = best score for this query + average index score
                # Limit max to 6 * average index score
                maxscore = min(scores[0][1] + self.avgscore, 6 * self.avgscore)

                # Normalize scores between 0 - 1 using maxscore
                scores = [(x, min(score / maxscore, 1.0)) for x, score in scores]

            # Add content, if available
            return self.results(scores)

        return None

    def batchsearch(self, queries, limit=3, threads=True):
        # Calculate number of threads using a thread per 25k records in index
        threads = math.ceil(self.count() / 25000) if isinstance(threads, bool) and threads else int(threads)
        threads = min(max(threads, 1), os.cpu_count())

        # This method is able to run as multiple threads due to a number of regex and numpy method calls that drop the GIL.
        results = []
        with ThreadPool(threads) as pool:
            for result in pool.starmap(self.search, [(x, limit) for x in queries]):
                results.append(result)

        return results

    def count(self):
        return self.terms.count() if self.terms else self.total

    def load(self, path):
        # Load scoring
        state = Serializer.load(path)

        # Convert to Counter instances
        for key in ["docfreq", "wordfreq", "tags"]:
            state[key] = Counter(state[key])

        # Convert documents to dict
        state["documents"] = dict(state["documents"]) if state["documents"] else state["documents"]

        # Set parameters on this object
        self.__dict__.update(state)

        # Load terms
        if self.config.get("terms"):
            self.terms = Terms(self.config["terms"], self.score, self.idf)
            self.terms.load(path + ".terms")

    def save(self, path):
        # Don't serialize following fields
        skipfields = ("config", "terms", "tokenizer")

        # Get object state
        state = {key: value for key, value in self.__dict__.items() if key not in skipfields}

        # Update documents to tuples
        state["documents"] = list(state["documents"].items()) if state["documents"] else state["documents"]

        # Save scoring
        Serializer.save(state, path)

        # Save terms
        if self.terms:
            self.terms.save(path + ".terms")

    def close(self):
        if self.terms:
            self.terms.close()

    def issparse(self):
        return self.terms is not None

    def isnormalized(self):
        return self.normalize

    def computefreq(self, tokens):
        """
        Computes token frequency. Used for token weighting.

        Args:
            tokens: input tokens

        Returns:
            {token: count}
        """

        return Counter(tokens)

    def computeidf(self, freq):
        """
        Computes an idf score for word frequency.

        Args:
            freq: word frequency

        Returns:
            idf score
        """

        return np.log((self.total + 1) / (freq + 1)) + 1

    # pylint: disable=W0613
    def score(self, freq, idf, length):
        """
        Calculates a score for each token.

        Args:
            freq: token frequency
            idf: token idf score
            length: total number of tokens in source document

        Returns:
            token score
        """

        return idf * np.sqrt(freq) * (1 / np.sqrt(length))

    def addstats(self, tokens, tags):
        """
        Add tokens and tags to stats.

        Args:
            tokens: list of tokens
            tags: list of tags
        """

        # Total number of times token appears, count all tokens
        self.wordfreq.update(tokens)

        # Total number of documents a token is in, count unique tokens
        self.docfreq.update(set(tokens))

        # Get list of unique tags
        if tags:
            self.tags.update(tags.split())

        # Total document count
        self.total += 1

    def tokenize(self, text):
        """
        Tokenizes text using default tokenizer.

        Args:
            text: input text

        Returns:
            tokens
        """

        # Load tokenizer
        if not self.tokenizer:
            self.tokenizer = self.loadtokenizer()

        return self.tokenizer(text)

    def loadtokenizer(self):
        """
        Load default tokenizer.

        Returns:
            tokenize method
        """

        # Custom tokenizer settings
        if self.config.get("tokenizer"):
            return Tokenizer(**self.config.get("tokenizer"))

        # Terms index use a standard tokenizer
        if self.config.get("terms"):
            return Tokenizer()

        # Standard scoring index without a terms index uses backwards compatible static tokenize method
        return Tokenizer.tokenize

    def results(self, scores):
        """
        Resolves a list of (id, score) with document content, if available. Otherwise, the original input is returned.

        Args:
            scores: list of (id, score)

        Returns:
            resolved results
        """

        # Convert to Python values
        scores = [(x, float(score)) for x, score in scores]

        if self.documents:
            return [{"id": x, "text": self.documents[x], "score": score} for x, score in scores]

        return scores



"""
Caption module
"""

# Conditional import
try:
    from PIL import Image

    PIL = True
except ImportError:
    PIL = False

from ..hfpipeline import HFPipeline


class Caption(HFPipeline):
    """
    Constructs captions for images.
    """

    def __init__(self, path=None, quantize=False, gpu=True, model=None, **kwargs):
        if not PIL:
            raise ImportError('Captions pipeline is not available - install "pipeline" extra to enable')

        # Call parent constructor
        super().__init__("image-to-text", path, quantize, gpu, model, **kwargs)

    def __call__(self, images):
        """
        Builds captions for images.

        This method supports a single image or a list of images. If the input is an image, the return
        type is a string. If text is a list, a list of strings is returned

        Args:
            images: image|list

        Returns:
            list of captions
        """

        # Convert single element to list
        values = [images] if not isinstance(images, list) else images

        # Open images if file strings
        values = [Image.open(image) if isinstance(image, str) else image for image in values]

        # Get and clean captions
        captions = []
        for result in self.pipeline(values):
            text = " ".join([r["generated_text"] for r in result]).strip()
            captions.append(text)

        # Return single element if single element passed in
        return captions[0] if not isinstance(images, list) else captions



"""
UrlTask module
"""

import re

from .base import Task


class UrlTask(Task):
    """
    Task that processes urls
    """

    # URL prefix
    PREFIX = r"\w+:\/\/"

    def accept(self, element):
        # Only accept elements that start with a url prefix
        return super().accept(element) and re.match(UrlTask.PREFIX, element.lower())



"""
Late encoder module
"""

import numpy as np
import torch

from ...models import Models, PoolingFactory
from ..base import Pipeline


class LateEncoder(Pipeline):
    """
    Computes similarity between query and list of text using a late interaction model.
    """

    def __init__(self, path=None, **kwargs):
        # Get device
        self.device = Models.device(Models.deviceid(kwargs.get("gpu", True)))

        # Load model
        self.model = PoolingFactory.create(
            {
                "method": kwargs.get("method"),
                "path": path if path else "colbert-ir/colbertv2.0",
                "device": self.device,
                "tokenizer": kwargs.get("tokenizer"),
                "maxlength": kwargs.get("maxlength"),
                "modelargs": {**kwargs.get("vectors", {}), **{"muvera": None}},
            }
        )

    def __call__(self, query, texts, limit=None):
        """
        Computes the similarity between query and list of text. Returns a list of
        (id, score) sorted by highest score, where id is the index in texts.

        This method supports query as a string or a list. If the input is a string,
        the return type is a 1D list of (id, score). If text is a list, a 2D list
        of (id, score) is returned with a row per string.

        Args:
            query: query text|list
            texts: list of text
            limit: maximum comparisons to return, defaults to all

        Returns:
            list of (id, score)
        """

        queries = [query] if isinstance(query, str) else query

        # Encode text to vectors
        queries = self.encode(queries, "query")
        data = self.encode(texts, "data") if isinstance(texts[0], str) else texts

        # Compute maximum similarity score
        scores = []
        for q in queries:
            scores.extend(self.score(q.unsqueeze(0), data, limit))

        return scores[0] if isinstance(query, str) else scores

    def encode(self, data, category):
        """
        Encodes a batch of data using the underlying model.

        Args:
            data: input data
            category: encoding category

        Returns:
            encoded data
        """

        return torch.from_numpy(self.model.encode(data, category=category)).to(self.device)

    def score(self, queries, data, limit):
        """
        Computes the maximum similarity score between query vectors and data vectors.

        Args:
            queries: query vectors
            data: data vectors
            limit: query limit

        Returns:
            list of (id, score)
        """

        # Compute bulk dot product using einstein notation
        scores = torch.einsum("ash,bth->abst", queries, data).max(axis=-1).values.mean(axis=-1)
        scores = scores.cpu().numpy()

        # Get top n matching indices and scores
        indices = np.argpartition(-scores, limit if limit and limit < scores.shape[0] else scores.shape[0] - 1)[:, :limit]
        scores = np.take_along_axis(scores, indices, axis=1)

        results = []
        for x, index in enumerate(indices):
            results.append(list(zip(index.tolist(), scores[x].tolist())))

        return results



"""
SparseVectors module
"""

# Conditional import
try:
    from scipy.sparse import csr_matrix, vstack
    from sklearn.preprocessing import normalize
    from sklearn.utils.extmath import safe_sparse_dot

    SPARSE = True
except ImportError:
    SPARSE = False

from ...util import SparseArray
from ..base import Vectors


# pylint: disable=W0223
class SparseVectors(Vectors):
    """
    Base class for sparse vector models. Vector models transform input content into sparse arrays.
    """

    def __init__(self, config, scoring, models):
        # Check before parent constructor since it calls loadmodel
        if not SPARSE:
            raise ImportError('SparseVectors is not available - install "vectors" extra to enable')

        super().__init__(config, scoring, models)

        # Get normalization setting
        self.isnormalize = self.config.get("normalize", self.defaultnormalize()) if self.config else None

    def encode(self, data, category=None):
        # Encode data to embeddings
        embeddings = super().encode(data, category)

        # Get sparse torch vector attributes
        embeddings = embeddings.cpu().coalesce()
        indices = embeddings.indices().numpy()
        values = embeddings.values().numpy()

        # Return as SciPy CSR Matrix
        return csr_matrix((values, indices), shape=embeddings.size())

    def vectors(self, documents, batchsize=500, checkpoint=None, buffer=None, dtype=None):
        # Run indexing
        ids, dimensions, batches, stream = self.index(documents, batchsize, checkpoint)

        # Rebuild sparse array
        embeddings = None
        with open(stream, "rb") as queue:
            for _ in range(batches):
                # Read in array batch
                data = self.loadembeddings(queue)
                embeddings = vstack((embeddings, data)) if embeddings is not None else data

        # Return sparse array
        return (ids, dimensions, embeddings)

    def dot(self, queries, data):
        return safe_sparse_dot(queries, data.T, dense_output=True).tolist()

    def loadembeddings(self, f):
        return SparseArray().load(f)

    def saveembeddings(self, f, embeddings):
        SparseArray().save(f, embeddings)

    def truncate(self, embeddings):
        raise ValueError("Truncate is not supported for sparse vectors")

    def normalize(self, embeddings):
        # Optionally normalize embeddings using method that supports sparse vectors
        return normalize(embeddings, copy=False) if self.isnormalize else embeddings

    def quantize(self, embeddings):
        raise ValueError("Quantize is not supported for sparse vectors")

    def defaultnormalize(self):
        """
        Returns the default normalization setting.

        Returns:
            default normalization setting
        """

        # Sparse vector embeddings typically perform better as unnormalized
        return False



"""
Documents module
"""

import os
import tempfile

from ...serialize import SerializeFactory


class Documents:
    """
    Streams documents to temporary storage. Allows queuing large volumes of content for later indexing.
    """

    def __init__(self):
        """
        Creates a new documents stream.
        """

        self.documents = None
        self.batch = 0
        self.size = 0

        # Pickle serialization - local temporary data
        self.serializer = SerializeFactory.create("pickle", allowpickle=True)

    def __len__(self):
        """
        Returns total number of queued documents.
        """

        return self.size

    def __iter__(self):
        """
        Streams all queued documents.
        """

        # Close streaming file
        self.documents.close()

        # Open stream file
        with open(self.documents.name, "rb") as queue:
            # Read each batch
            for _ in range(self.batch):
                documents = self.serializer.loadstream(queue)

                # Yield each document
                yield from documents

    def add(self, documents):
        """
        Adds a batch of documents for indexing.

        Args:
            documents: list of (id, data, tag) tuples

        Returns:
            documents
        """

        # Create documents file if not already open
        # pylint: disable=R1732
        if not self.documents:
            self.documents = tempfile.NamedTemporaryFile(mode="wb", suffix=".docs", delete=False)

        # Add batch
        self.serializer.savestream(documents, self.documents)
        self.batch += 1
        self.size += len(documents)

        return documents

    def close(self):
        """
        Closes and resets this instance. New sets of documents can be added with additional calls to add.
        """

        # Cleanup stream file
        os.remove(self.documents.name)

        # Reset document parameters
        self.documents = None
        self.batch = 0
        self.size = 0



"""
MessagePack module
"""

import msgpack
from msgpack import Unpacker
from msgpack.exceptions import ExtraData

from .base import Serialize
from .errors import SerializeError


class MessagePack(Serialize):
    """
    MessagePack serialization.
    """

    def __init__(self, streaming=False, **kwargs):
        # Parent constructor
        super().__init__()

        # Streaming unpacker
        self.streaming = streaming

        # Additional streaming unpacker keyword arguments
        self.kwargs = kwargs

    def loadstream(self, stream):
        try:
            # Support both streaming and non-streaming unpacking of data
            return Unpacker(stream, **self.kwargs) if self.streaming else msgpack.unpack(stream)
        except ExtraData as e:
            raise SerializeError(e) from e

    def savestream(self, data, stream):
        msgpack.pack(data, stream)

    def loadbytes(self, data):
        return msgpack.unpackb(data)

    def savebytes(self, data):
        return msgpack.packb(data)



"""
BM25 module
"""

import numpy as np

from .tfidf import TFIDF


class BM25(TFIDF):
    """
    Best matching (BM25) scoring.
    """

    def __init__(self, config=None):
        super().__init__(config)

        # BM25 configurable parameters
        self.k1 = self.config.get("k1", 1.2)
        self.b = self.config.get("b", 0.75)

    def computeidf(self, freq):
        # Calculate BM25 IDF score
        return np.log(1 + (self.total - freq + 0.5) / (freq + 0.5))

    def score(self, freq, idf, length):
        # Calculate BM25 score
        k = self.k1 * ((1 - self.b) + self.b * length / self.avgdl)
        return idf * (freq * (self.k1 + 1)) / (freq + k)



"""
JSON module
"""

import base64
import json

from io import BytesIO
from typing import Any

import fastapi.responses

from PIL.Image import Image


class JSONEncoder(json.JSONEncoder):
    """
    Extended JSONEncoder that serializes images and byte streams as base64 encoded text.
    """

    def default(self, o):
        # Convert Image to BytesIO
        if isinstance(o, Image):
            buffered = BytesIO()
            o.save(buffered, format=o.format, quality="keep")
            o = buffered

        # Unpack bytes from BytesIO
        if isinstance(o, BytesIO):
            o = o.getvalue()

        # Base64 encode bytes instances
        if isinstance(o, bytes):
            return base64.b64encode(o).decode("utf-8")

        # Default handler
        return super().default(o)


class JSONResponse(fastapi.responses.JSONResponse):
    """
    Extended JSONResponse that serializes images and byte streams as base64 encoded text.
    """

    def render(self, content: Any) -> bytes:
        """
        Renders content to the response.

        Args:
            content: input content

        Returns:
            rendered content as bytes
        """

        return json.dumps(content, ensure_ascii=False, allow_nan=False, indent=None, separators=(",", ":"), cls=JSONEncoder).encode("utf-8")



"""
Scoring imports
"""

from .base import Scoring
from .bm25 import BM25
from .factory import ScoringFactory
from .pgtext import PGText
from .sif import SIF
from .sparse import Sparse
from .terms import Terms
from .tfidf import TFIDF



"""
HTMLToMarkdown module
"""

import re

# Conditional import
try:
    from bs4 import BeautifulSoup, NavigableString

    SOUP = True
except ImportError:
    SOUP = False

from ..base import Pipeline


class HTMLToMarkdown(Pipeline):
    """
    HTML to Markdown pipeline.

    Markdown formatting is applied for headings, blockquotes, lists, code, tables and text. Visual formatting is also
    included (bold, italic etc).

    This pipeline searches for the best node that has relevant text, often found with an article, main or body tag.
    """

    def __init__(self, paragraphs=False, sections=False):
        """
        Create a new Extract instance.

        Args:
            paragraphs: True if paragraph parsing enabled, False otherwise
            sections: True if section parsing enabled, False otherwise
        """

        if not SOUP:
            raise ImportError('HTMLToMarkdown pipeline is not available - install "pipeline" extra to enable')

        self.paragraphs = paragraphs
        self.sections = sections

    def __call__(self, html):
        """
        Transforms input HTML into Markdown formatted text.

        Args:
            html: input html

        Returns:
            markdown formatted text
        """

        # HTML Parser
        soup = BeautifulSoup(html, features="html.parser")

        # Ignore script and style tags
        for script in soup.find_all(["script", "style"]):
            script.decompose()

        # Check for article sections
        article = next((x for x in ["article", "main"] if soup.find(x)), None)

        # Extract text from each section element
        nodes = []
        for node in soup.find_all(article if article else "body"):
            # Skip article sections without at least 1 paragraph
            if not article or node.find("p"):
                nodes.append(self.process(node, article))

        # Return extracted text, fallback to default text extraction if no nodes found
        return "\n".join(self.metadata(soup) + nodes) if nodes else self.default(soup)

    def process(self, node, article):
        """
        Extracts text from a node. This method applies transforms for headings, blockquotes, lists, code, tables and text.
        Page breaks are detected and reflected in the output text as a page break character.

        Args:
            node: input node
            article: True if the main section node is an article

        Returns:
            node text
        """

        if self.isheader(node):
            return self.header(node, article)

        if node.name in ("blockquote", "q"):
            return self.block(node)

        if node.name in ("ul", "ol"):
            return self.items(node, article)

        if node.name in ("code", "pre"):
            return self.code(node)

        if node.name == "table":
            return self.table(node, article)

        # Nodes to skip
        if node.name in ("aside",) + (() if article else ("header", "footer")):
            return ""

        # Get page break symbol, if available
        page = node.name and node.get("class") and "page" in node.get("class")

        # Get node children
        children = self.children(node)

        # Join elements into text
        if self.iscontainer(node, children):
            texts = [self.process(node, article) for node in children]
            text = "\n".join(text for text in texts if text or not article)
        else:
            text = self.text(node, article)

        # Add page breaks, if section parsing enabled. Otherwise add node text.
        return f"{text}\f" if page and self.sections else text

    def metadata(self, node):
        """
        Builds a metadata section. The metadata section consists of the title and
        description fields.

        Args:
            node: input document node

        Returns:
            metadata as a list
        """

        title = node.find("title")
        metadata = [f"**{title.text.strip()}**"] if title and title.text else []

        description = node.find("meta", attrs={"name": "description"})
        if description and description["content"]:
            metadata.append(f"\n*{description['content'].strip()}*")

        # Add separator
        if metadata:
            metadata.append("\f" if self.sections else "\n\n")

        return metadata

    def default(self, soup):
        """
        Default text handler when valid HTML isn't detected.

        Args:
            soup: BeautifulSoup object

        Returns:
            text
        """

        lines = []
        for line in soup.get_text().split("\n"):
            # Detect markdown headings and add page breaks
            lines.append(f"\f{line}" if self.sections and re.search(r"^#+ ", line) else line)

        return "\n".join(lines)

    def text(self, node, article):
        """
        Text handler. This method flattens a node and it's children to text.

        Args:
            node: input node
            article: True if the main section node is an article

        Returns:
            node text
        """

        # Get node children if available, otherwise use node as item
        items = self.children(node)
        items = items if items else [node]

        # Apply emphasis and link formatting
        texts = []
        for x in items:
            target, text = x if x.name else node, x.text

            if text.strip():
                if target.name in ("b", "strong"):
                    text = f"**{text.strip()}** "
                elif target.name in ("i", "em"):
                    text = f"*{text.strip()}* "
                elif target.name == "a":
                    text = f"[{text.strip()}]({target.get('href')}) "

            texts.append(text)

        # Join text elements
        text = "".join(texts)

        # Article text processing
        text = self.articletext(node, text) if article else text

        # Return text, strip leading/trailing whitespace if this is a string only node
        text = text if node.name and text else text.strip()

        return text

    def header(self, node, article):
        """
        Header handler. This method transforms a HTML heading into a Markdown formatted heading.

        Args:
            node: input node
            article: True if the main section node is an article

        Returns:
            heading as markdown
        """

        # Get heading level and text
        level = "#" * int(node.name[1])
        text = self.text(node, article)

        # Add section break or newline, if necessary
        level = f"\f{level}" if self.sections else f"\n{level}"

        # Return formatted header. Remove leading whitespace as it was added before level in step above.
        return f"{level} {text.lstrip()}" if text.strip() else ""

    def block(self, node):
        """
        Blockquote handler. This method transforms a HTML blockquote or q block into a Markdown formatted
        blockquote

        Args:
            node: input node

        Returns:
            block as markdown
        """

        text = "\n".join(f"> {x}" for x in node.text.strip().split("\n"))
        return f"{text}\n\n" if self.paragraphs else f"{text}\n"

    def items(self, node, article):
        """
        List handler. This method transforms a HTML ordered/unordered list into a Markdown formatted list.

        Args:
            node: input node
            article: True if the main section node is an article

        Returns:
            list as markdown
        """

        elements = []
        for x, element in enumerate(node.find_all("li")):
            # Unordered lists use dashes. Ordered lists use numbers.
            prefix = "-" if node.name == "ul" else f"{x + 1}."

            # List item text
            text = self.process(element, article)

            # Add list element
            if text:
                elements.append(f"{prefix} {text}")

        # Join elements together as string
        return "\n".join(elements)

    def code(self, node):
        """
        Code block handler. This method transforms a HTML pre or code block into a Markdown formatted
        code block.

        Args:
            node: input node

        Returns:
            code as markdown
        """

        text = f"```\n{node.text.strip()}\n```"
        return f"{text}\n\n" if self.paragraphs else f"{text}\n"

    def table(self, node, article):
        """
        Table handler. This method transforms a HTML table into a Markdown formatted table.

        Args:
            node: input node
            article: True if the main section node is an article

        Returns:
            table as markdown
        """

        elements, header = [], False

        # Process all rows
        rows = node.find_all("tr")
        for row in rows:
            # Get list of columns for row
            columns = row.find_all(lambda tag: tag.name in ("th", "td"))

            # Add columns with separator
            elements.append(f"|{'|'.join(self.process(column, article) for column in columns)}|")

            # If there are multiple rows, add header format row
            if not header and len(rows) > 1:
                elements.append(f"{'|---' * len(columns)}|")
                header = True

        # Join elements together as string
        return "\n".join(elements)

    def iscontainer(self, node, children):
        """
        Analyzes a node and it's children to determine if this is a container element. A container
        element is defined as being a div, body, article or not having any string elements as children.

        Args:
            node: input node
            nodes: input node's children

        Returns:
            True if this is a container element, False otherwise
        """

        return children and (node.name in ("div", "body", "article") or not any(isinstance(x, NavigableString) for x in children))

    def children(self, node):
        """
        Gets the node children, if available.

        Args:
            node: input node

        Returns:
            node children or None if not available
        """

        if node.name and node.contents:
            # Iterate over children and remove whitespace-only string nodes
            return [node for node in node.contents if node.name or node.text.strip()]

        return None

    def articletext(self, node, text):
        """
        Transforms node text using article parsing rules. Article parsing is designed to extract text content from web articles.
        It ignores navigation headers and other superfluous elements.

        Args:
            node: input node
            text: current text

        Returns:
            article text
        """

        # List of valid text nodes
        valid = ("p", "th", "td", "li", "a", "b", "strong", "i", "em")

        # Check if this node is valid or it's part of a table cell
        valid = node.name in valid or (node.parent and node.parent.name in ("th", "td"))

        # Check if text is valid article text
        text = text if (valid or self.isheader(node)) and not self.islink(node) else ""
        if text:
            # Replace non-breaking space plus newline with double newline
            text = text.replace("\xa0\n", "\n\n")

            # Format paragraph whitespace
            if node.name == "p":
                text = f"{text.strip()}\n\n" if self.paragraphs else f"{text.strip()}\n"

        return text

    def isheader(self, node):
        """
        Checks if node is a header node.

        Args:
            node: input node

        Returns:
            True if node is a header node, False otherwise
        """

        return node.name in ("h1", "h2", "h3", "h4", "h5", "h6")

    def islink(self, node):
        """
        Checks if node is a link node. This method does not consider links without tables as link nodes.

        Args:
            node: input node

        Returns:
            True if node is a link node, False otherwise
        """

        # Check if this is a link node or link container
        link, parent = False, node
        while parent:
            if parent.name == "a":
                link = True
                break

            parent = parent.parent

        # Return if this node or any parents are a link. Ignore links in table cells.
        return link and node.parent.name not in ("th", "td")



"""
Dense ANN imports
"""

from .annoy import Annoy
from .factory import ANNFactory
from .faiss import Faiss
from .hnsw import HNSW
from .numpy import NumPy
from .pgvector import PGVector
from .torch import Torch



"""
ImageEncoder module
"""

from io import BytesIO

# Conditional import
try:
    from PIL import Image

    PIL = True
except ImportError:
    PIL = False

from .base import Encoder


class ImageEncoder(Encoder):
    """
    Encodes and decodes Image objects as compressed binary content, using the original image's algorithm.
    """

    def __init__(self):
        """
        Creates a new ImageEncoder.
        """

        if not PIL:
            raise ImportError('ImageEncoder is not available - install "database" extra to enable')

    def encode(self, obj):
        # Create byte stream
        output = BytesIO()

        # Write image to byte stream
        obj.save(output, format=obj.format, quality="keep")

        # Return byte array
        return output.getvalue()

    def decode(self, data):
        # Return a PIL image
        return Image.open(BytesIO(data)) if data else None



"""
Factory module
"""

from urllib.parse import urlparse

from ..util import Resolver

from .client import Client
from .duckdb import DuckDB
from .sqlite import SQLite


class DatabaseFactory:
    """
    Methods to create document databases.
    """

    @staticmethod
    def create(config):
        """
        Create a Database.

        Args:
            config: database configuration parameters

        Returns:
            Database
        """

        # Database instance
        database = None

        # Enables document database
        content = config.get("content")

        # Standardize content name
        if content is True:
            content = "sqlite"

        # Create document database instance
        if content == "duckdb":
            database = DuckDB(config)
        elif content == "sqlite":
            database = SQLite(config)
        elif content:
            # Check if content is a URL
            url = urlparse(content)
            if content == "client" or url.scheme:
                # Connect to database server URL
                database = Client(config)
            else:
                # Resolve custom database if content is not a URL
                database = DatabaseFactory.resolve(content, config)

        # Store config back
        config["content"] = content

        return database

    @staticmethod
    def resolve(backend, config):
        """
        Attempt to resolve a custom backend.

        Args:
            backend: backend class
            config: index configuration parameters

        Returns:
            Database
        """

        try:
            return Resolver()(backend)(config)
        except Exception as e:
            raise ImportError(f"Unable to resolve database backend: '{backend}'") from e



"""
Hugging Face Hub module
"""

import os
import tempfile

import huggingface_hub

from huggingface_hub.utils import RepositoryNotFoundError

from .base import Cloud


class HuggingFaceHub(Cloud):
    """
    Hugging Face Hub cloud provider.
    """

    def metadata(self, path=None):
        try:
            # If this is an archive file path, get file metadata
            if self.isarchive(path):
                url = huggingface_hub.hf_hub_url(
                    repo_id=self.config["container"], filename=os.path.basename(path), revision=self.config.get("revision")
                )

                return huggingface_hub.get_hf_file_metadata(url=url, token=self.config.get("token"))

            # Otherwise return repository metadata
            return huggingface_hub.model_info(repo_id=self.config["container"], revision=self.config.get("revision"), token=self.config.get("token"))

        except RepositoryNotFoundError:
            return None

    def load(self, path=None):
        # Download archvie file and return local path
        if self.isarchive(path):
            return huggingface_hub.hf_hub_download(
                repo_id=self.config["container"],
                filename=os.path.basename(path),
                revision=self.config.get("revision"),
                cache_dir=self.config.get("cache"),
                token=self.config.get("token"),
            )

        # Download repository and return cached path
        return huggingface_hub.snapshot_download(
            repo_id=self.config["container"], revision=self.config.get("revision"), cache_dir=self.config.get("cache"), token=self.config.get("token")
        )

    def save(self, path):
        # Get or create repository
        huggingface_hub.create_repo(
            repo_id=self.config["container"], token=self.config.get("token"), private=self.config.get("private", True), exist_ok=True
        )

        # Enable lfs-tracking of embeddings index files
        self.lfstrack()

        # Upload files
        for f in self.listfiles(path):
            huggingface_hub.upload_file(
                repo_id=self.config["container"],
                revision=self.config.get("revision"),
                token=self.config.get("token"),
                path_or_fileobj=f,
                path_in_repo=os.path.basename(f),
            )

    def lfstrack(self):
        """
        Adds lfs-tracking of embeddings index files. This method adds tracking for documents and embeddings to .gitattributes.
        """

        # Get and read .gitattributes file
        path = huggingface_hub.hf_hub_download(
            repo_id=self.config["container"], filename=os.path.basename(".gitattributes"), token=self.config.get("token")
        )

        with open(path, "r", encoding="utf-8") as f:
            content = f.read()

        # Check if index files are lfs-tracked. Update .gitattributes, if necessary.
        if "embeddings " not in content:
            # Add documents and embeddings to lfs tracking
            content += "documents filter=lfs diff=lfs merge=lfs -text\n"
            content += "embeddings filter=lfs diff=lfs merge=lfs -text\n"

            # pylint: disable=R1732
            with tempfile.NamedTemporaryFile(mode="w", delete=False) as tmp:
                tmp.write(content)
                attributes = tmp.name

            # Upload file
            huggingface_hub.upload_file(
                repo_id=self.config["container"], token=self.config.get("token"), path_or_fileobj=attributes, path_in_repo=os.path.basename(path)
            )

            # Remove temporary file
            os.remove(attributes)



"""
SQLite module
"""

import os
import sqlite3

from .embedded import Embedded


class SQLite(Embedded):
    """
    Database instance backed by SQLite.
    """

    def connect(self, path=""):
        # Create connection
        connection = sqlite3.connect(path, check_same_thread=False)

        # Enable WAL mode, if necessary
        if self.setting("wal"):
            connection.execute("PRAGMA journal_mode=WAL")

        return connection

    def getcursor(self):
        return self.connection.cursor()

    def rows(self):
        return self.cursor

    def addfunctions(self):
        if self.connection and self.functions:
            # Enable callback tracebacks to show user-defined function errors
            sqlite3.enable_callback_tracebacks(True)

            for name, argcount, fn in self.functions:
                self.connection.create_function(name, argcount, fn)

    def copy(self, path):
        # Delete existing file, if necessary
        if os.path.exists(path):
            os.remove(path)

        # Create database. Thread locking must be handled externally.
        connection = self.connect(path)

        if self.connection.in_transaction:
            # The backup call will hang if there are uncommitted changes, need to copy over
            # with iterdump (which is much slower)
            for sql in self.connection.iterdump():
                connection.execute(sql)
        else:
            # Database is up to date, can do a more efficient copy with SQLite C API
            self.connection.backup(connection)

        return connection
